from jolt import Task
from jolt import utils
from jolt.error import raise_task_error_if

import os


class CMake(Task):
    """ Builds and publishes a CMake project """

    abstract = True

    cmakelists = "CMakeLists.txt"
    """ Path to CMakeLists.txt or directory containing CMakelists.txt """

    generator = None

    incremental = True

    options = []
    """ List of options and their values (``option[:type]=value``) """

    def run(self, deps, tools):
        raise_task_error_if(not self.cmakelists, self, "cmakelists attribute has not been defined")

        cmake = tools.cmake(incremental=self.incremental)
        cmake.configure(tools.expand(self.cmakelists), *["-D" + tools.expand(option) for option in self.options], generator=utils.quote(self.generator, "'"))
        cmake.build()
        cmake.install()

    def publish(self, artifact, tools):
        cmake = tools.cmake()
        cmake.publish(artifact)


class _CMakeCXXLibrary(CMake):
    binary = None

    generator = "Eclipse CDT4 - Ninja"

    incremental = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.binary = self.binary or self.canonical_name

    def run(self, deps, tools):
        with tools.cwd(tools.builddir(incremental=self.incremental)):
            self.cmakelists = tools.expand_path(self.cmakelists)
            project = utils.render("cxxlibrary.cmake.template", deps=deps, task=self, tools=tools, os=os)
            print(project)
            tools.write_file(self.cmakelists, project)
        super().run(deps, tools)

    def publish(self, artifact, tools):
        super().publish(artifact, tools)



import functools
import sys

from jolt import cache
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.hooks import TaskHook, TaskHookFactory
from jolt.influence import StringInfluence
from jolt.plugins import ninja
from jolt.plugins import ninjacli

log.verbose("[NinjaCache] Loaded")


class CacheHooks(TaskHook):
    def task_created(self, task):
        if not isinstance(task.task, ninja.CXXLibrary) or task.task.shared:
            return
        task.task.influence.append(StringInfluence("NinjaCache"))
        task.task._write_ninja_cache = functools.partial(self.task_post_ninja_file, task)

    def task_post_ninja_file(self, task, deps, tools):
        if not isinstance(task.task, ninja.CXXLibrary) or task.task.shared:
            return

        cli = fs.path.join(fs.path.dirname(__file__), "ninjacli.py")
        disabled = config.getboolean("ninja-cache", "disable", False)

        tools.setenv("CCWRAP", "{} {} -- ".format(sys.executable, cli))
        tools.setenv("CXXWRAP", "{} {} -- ".format(sys.executable, cli))
        tools.setenv("JOLT_CACHEDIR", cache.ArtifactCache.get().root)
        tools.setenv("JOLT_CANONTASK", utils.canonical(task.task.name))
        tools.setenv("NINJACACHE_DISABLE", "1" if disabled else "0")
        tools.setenv("NINJACACHE_MAXARTIFACTS", config.getint("ninja-cache", "maxartifacts", 0))
        if log.is_verbose():
            tools.setenv("NINJACACHE_VERBOSE", "1")

        if not disabled:
            objcache = ninjacli.Cache(tools.builddir("ninja", task.task.incremental))
            objcache.load_manifests(tools.getenv("JOLT_CACHEDIR"), tools.getenv("JOLT_CANONTASK"))
            objcache.save()

    def task_postpublish(self, task, artifact, tools):
        if not isinstance(task.task, ninja.CXXLibrary) or task.task.shared:
            return
        with tools.cwd(task.task.outdir):
            m = ninjacli.LibraryManifest(tools.expand_path(".ninja.json"))
            m.read()
            if hasattr(artifact.strings, "library"):
                m.add_library(fs.path.join(
                    task.task.publishdir,
                    fs.path.basename(str(artifact.strings.library))))
            m.write()
            artifact.collect(".ninja.json")


@TaskHookFactory.register
class CacheHookFactory(TaskHookFactory):
    def create(self, env):
        return CacheHooks()



import click
import functools
import getpass
import keyring
import os
import pika
from pika.exceptions import AMQPConnectionError
from pika.adapters.utils.connection_workflow import AMQPConnectorStackTimeout
import threading
import time

from jolt.cli import cli
from jolt import config
from jolt import filesystem as fs
from jolt import hooks
from jolt import log
from jolt import scheduler
from jolt import utils
from jolt.manifest import JoltManifest
from jolt.tools import Tools
from jolt.error import JoltCommandError
from jolt.error import raise_error
from jolt.error import raise_task_error_if
from jolt.error import raise_task_error_on_exception


NAME = "amqp"


def _get_auth():
    service = config.get(NAME, "keyring.service", NAME)

    username = config.get(NAME, "keyring.username")
    if not username:
        return "guest", "guest"

    password = config.get(NAME, "keyring.password") or keyring.get_password(NAME, username)
    if not password:
        password = getpass.getpass(NAME + " password: ")
        assert password, "no password in keyring for " + NAME
        keyring.set_password(service, username, password)
    return username, password


def _get_url():
    host = config.get("amqp", "host", "amqp-service")
    port = int(config.get("amqp", "port", 5672))
    username, password = _get_auth()
    return "amqp://{}:{}@{}:{}/%2F".format(username, password, host, port)


TYPE = "Remote execution"
TIMEOUT = (3.5, 27)
POLL_INTERVAL = 1


class WorkerTaskConsumer(object):
    """This is a task consumer that will handle unexpected interactions
    with RabbitMQ such as channel and connection closures.

    If RabbitMQ closes the connection, this class will stop and indicate
    that reconnection is necessary. You should look at the output, as
    there are limited reasons why the connection may be closed, which
    usually are tied to permission related issues or socket timeouts.

    If the channel is closed, it will indicate a problem with one of the
    commands that were issued and that should surface in the output as well.

    """
    EXCHANGE = 'jolt_exchange'
    EXCHANGE_TYPE = 'direct'
    QUEUE = 'jolt_tasks'
    RESULT_EXCHANGE = 'jolt_results'
    ROUTING_KEY_PREFIX = ''
    ROUTING_KEY_REQUEST = 'default'
    ROUTING_KEY_RESULT = 'default'

    def __init__(self, amqp_url):
        """Create a new instance of the consumer class, passing in the AMQP
        URL used to connect to RabbitMQ.

        :param str amqp_url: The AMQP url to connect with

        """
        self.should_reconnect = False
        self.was_consuming = False

        self._connection = None
        self._channel = None
        self._closing = False
        self._consumer_tag = None
        self._url = amqp_url
        self._consuming = False
        # In production, experiment with higher prefetch values
        # for higher consumer throughput
        self._prefetch_count = 1
        self._job = None
        self._routing_key = self.ROUTING_KEY_PREFIX + config.get(
            "amqp", "routing_key",
            os.getenv("RABBITMQ_ROUTING_KEY", self.ROUTING_KEY_REQUEST))
        self._queue = self.QUEUE + "_" + self._routing_key
        self._max_priority = config.getint("amqp", "max-priority", 1)

    def connect(self):
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.

        :rtype: pika.SelectConnection

        """
        log.info('Connecting to {}', self._url)
        return pika.SelectConnection(
            parameters=pika.URLParameters(self._url),
            on_open_callback=self.on_connection_open,
            on_open_error_callback=self.on_connection_open_error,
            on_close_callback=self.on_connection_closed)

    def close_connection(self):
        self._consuming = False
        if self._connection.is_closing or self._connection.is_closed:
            log.info('Connection is closing or already closed')
        else:
            log.info('Closing connection')
            self._connection.close()

    def on_connection_open(self, _unused_connection):
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.

        :param pika.SelectConnection _unused_connection: The connection

        """
        log.info('Connection opened')
        self.open_channel()

    def on_connection_open_error(self, _unused_connection, err):
        """This method is called by pika if the connection to RabbitMQ
        can't be established.

        :param pika.SelectConnection _unused_connection: The connection
        :param Exception err: The error

        """
        log.error('Connection open failed: {}', err)
        self.reconnect()

    def on_connection_closed(self, _unused_connection, reason):
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        :param pika.connection.Connection connection: The closed connection obj
        :param Exception reason: exception representing reason for loss of
            connection.

        """
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            log.warning('Connection closed, reconnect necessary: {}', reason)
            self.reconnect()

    def reconnect(self):
        """Will be invoked if the connection can't be opened or is
        closed. Indicates that a reconnect is necessary then stops the
        ioloop.

        """
        self.should_reconnect = True
        self.stop()

    def open_channel(self):
        """Open a new channel with RabbitMQ by issuing the Channel.Open RPC
        command. When RabbitMQ responds that the channel is open, the
        on_channel_open callback will be invoked by pika.

        """
        log.info('Creating a new channel')
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.

        Since the channel is now open, we'll declare the exchange to use.

        :param pika.channel.Channel channel: The channel object

        """
        log.info('Channel opened')
        self._channel = channel
        self.add_on_channel_close_callback()
        self.setup_exchange(self.EXCHANGE)

    def add_on_channel_close_callback(self):
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.

        """
        log.info('Adding channel close callback')
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reason):
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        :param pika.channel.Channel: The closed channel
        :param Exception reason: why the channel was closed

        """
        log.warning('Channel {} was closed: {}', channel, reason)
        self.close_connection()

    def setup_exchange(self, exchange_name):
        """Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.

        :param str|unicode exchange_name: The name of the exchange to declare

        """
        log.info('Declaring exchange: {}', exchange_name)
        # Note: using functools.partial is not required, it is demonstrating
        # how arbitrary data can be passed to the callback when it is called
        cb = functools.partial(
            self.on_exchange_declareok, userdata=exchange_name)
        self._channel.exchange_declare(
            exchange=exchange_name,
            exchange_type=self.EXCHANGE_TYPE,
            callback=cb)

    def on_exchange_declareok(self, _unused_frame, userdata):
        """Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
        command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame
        :param str|unicode userdata: Extra user data (exchange name)

        """
        log.info('Exchange declared: {}', userdata)
        self.setup_queue(self._queue)

    def setup_queue(self, queue_name):
        """Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command. When it is complete, the on_queue_declareok method will
        be invoked by pika.

        :param str|unicode queue_name: The name of the queue to declare.

        """
        log.info('Declaring queue {}', queue_name)
        cb = functools.partial(self.on_queue_declareok, userdata=queue_name)
        self._channel.queue_declare(
            queue=queue_name, callback=cb,
            arguments={
                "x-message-deduplication": True,
                "x-max-priority": self._max_priority
            })

    def on_queue_declareok(self, _unused_frame, userdata):
        """Method invoked by pika when the Queue.Declare RPC call made in
        setup_queue has completed. In this method we will bind the queue
        and exchange together with the routing key by issuing the Queue.Bind
        RPC command. When this command is complete, the on_bindok method will
        be invoked by pika.

        :param pika.frame.Method _unused_frame: The Queue.DeclareOk frame
        :param str|unicode userdata: Extra user data (queue name)

        """
        queue_name = userdata
        log.info('Binding {} to {} with {}',
                 self.EXCHANGE, queue_name, self._routing_key)
        cb = functools.partial(self.on_bindok, userdata=queue_name)
        self._channel.queue_bind(
            queue_name,
            self.EXCHANGE,
            routing_key=self._routing_key,
            callback=cb)
        log.info('Binding {} to {} with jolt_{}',
                 self.EXCHANGE, queue_name, self._routing_key)
        self._channel.queue_bind(
            queue_name,
            self.EXCHANGE,
            routing_key="jolt_" + self._routing_key)

    def on_bindok(self, _unused_frame, userdata):
        """Invoked by pika when the Queue.Bind method has completed. At this
        point we will set the prefetch count for the channel.

        :param pika.frame.Method _unused_frame: The Queue.BindOk response frame
        :param str|unicode userdata: Extra user data (queue name)

        """
        log.info('Queue bound: {}', userdata)
        self.set_qos()

    def set_qos(self):
        """This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before RabbitMQ will deliver another one. You should experiment
        with different prefetch values to achieve desired performance.

        """
        self._channel.basic_qos(
            prefetch_count=self._prefetch_count, callback=self.on_basic_qos_ok)

    def on_basic_qos_ok(self, _unused_frame):
        """Invoked by pika when the Basic.QoS method has completed. At this
        point we will start consuming messages by calling start_consuming
        which will invoke the needed RPC commands to start the process.

        :param pika.frame.Method _unused_frame: The Basic.QosOk response frame

        """
        log.info('QOS set to: {}', self._prefetch_count)
        self.start_consuming()

    def start_consuming(self):
        """This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if RabbitMQ
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with RabbitMQ. We keep the value to use it when we want to
        cancel consuming. The on_message method is passed in as a callback pika
        will invoke when a message is fully received.

        """
        log.info('Issuing consumer related RPC commands')
        self.add_on_cancel_callback()
        self._consumer_tag = self._channel.basic_consume(
            self._queue, self.on_message)
        self.was_consuming = True
        self._consuming = True

    def add_on_cancel_callback(self):
        """Add a callback that will be invoked if RabbitMQ cancels the consumer
        for some reason. If RabbitMQ does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.

        """
        log.info('Adding consumer cancellation callback')
        self._channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame

        """
        log.info('Consumer was cancelled remotely, shutting down: {}',
                 method_frame)
        if self._channel:
            self._channel.close()

    def on_message(self, channel, basic_deliver, properties, body):
        """Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.

        :param pika.channel.Channel _unused_channel: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param bytes body: The message body

        """
        log.info('Received execution request # {} from {}',
                 basic_deliver.delivery_tag, properties.app_id)

        class Job(threading.Thread):
            def __init__(self, consumer, channel, basic_deliver, properties, body):
                super(Job, self).__init__()
                self.consumer = consumer
                self.channel = channel
                self.basic_deliver = basic_deliver
                self.properties = properties
                self.body = body

            def selfdeploy(self):
                """ Installs the correct version of Jolt as specified in execution request. """

                tools = Tools()
                manifest = JoltManifest()
                try:
                    manifest.parse()
                    ident = manifest.get_parameter("jolt_identity")
                    url = manifest.get_parameter("jolt_url")
                    if not ident or not url:
                        return "jolt"

                    requires = manifest.get_parameter("jolt_requires")

                    log.info("Jolt version: {}", ident)

                    src = "build/selfdeploy/{}/src".format(ident)
                    env = "build/selfdeploy/{}/env".format(ident)

                    if not fs.path.exists(env):
                        try:
                            fs.makedirs(src)
                            tools.run("curl {} | tar zx -C {}", url, src)
                            tools.run("virtualenv {}", env)
                            tools.run(". {}/bin/activate && pip install -e {}", env, src)
                            if requires:
                                tools.run(". {}/bin/activate && pip install {}", env, requires)
                            if "autocompletion=" in tools.read_file(f"{src}/jolt/cli.py"):
                                tools.run(". {}/bin/activate && pip install 'click<8.1'", env, src)
                        except Exception as e:
                            tools.rmtree("build/selfdeploy/{}", ident, ignore_errors=True)
                            raise e

                    return ". {}/bin/activate && jolt".format(env)
                except Exception as e:
                    log.exception()
                    raise e

            def run(self):
                with open("default.joltxmanifest", "wb") as f:
                    f.write(self.body)

                log.info("Manifest written")

                tools = Tools()
                for recipe in tools.glob("*.jolt"):
                    tools.unlink(recipe)

                try:
                    jolt = self.selfdeploy()
                    config_file = config.get("amqp", "config", "")
                    if config_file:
                        config_file = "-c " + config_file

                    log.info("Running jolt")
                    tools.run("{} -vv {} build --worker --result result.joltxmanifest",
                              jolt, config_file, output_stdio=True)
                except JoltCommandError as e:
                    self.response = ""
                    try:
                        manifest = JoltManifest()
                        try:
                            manifest.parse("result.joltxmanifest")
                        except Exception:
                            manifest.duration = "0"
                        manifest.result = "FAILED"
                        manifest.stdout = "\n".join(e.stdout)
                        manifest.stderr = "\n".join(e.stderr)
                        self.response = manifest.format()
                    except Exception:
                        log.exception()
                    log.error("Task failed")
                except Exception:
                    log.exception()
                    self.response = ""
                    try:
                        manifest = JoltManifest()
                        try:
                            manifest.parse("result.joltxmanifest")
                        except Exception:
                            manifest.duration = "0"
                        manifest.result = "FAILED"
                        self.response = manifest.format()
                    except Exception:
                        log.exception()
                    log.error("Task failed")
                else:
                    self.response = ""
                    try:
                        manifest = JoltManifest()
                        try:
                            manifest.parse("result.joltxmanifest")
                        except Exception:
                            manifest.duration = "0"
                        manifest.result = "SUCCESS"
                        self.response = manifest.format()
                    except Exception:
                        log.exception()
                    log.info("Task succeeded")

                utils.call_and_catch(tools.unlink, "result.joltxmanifest")
                self.consumer.add_on_job_completed_callback(self)

        self._job = Job(self, channel, basic_deliver, properties, body)
        self._job.start()

    def add_on_job_completed_callback(self, job):
        if self._connection:
            self._connection.ioloop.add_callback_threadsafe(
                functools.partial(self.on_job_completed, job))
        else:
            self._job = None

    def on_job_completed(self, job):
        job.join()
        if job.channel is self._channel:
            self._channel.basic_publish(
                exchange=self.RESULT_EXCHANGE,
                routing_key=job.properties.correlation_id,
                properties=pika.BasicProperties(
                    correlation_id=job.properties.correlation_id,
                    expiration="600000"),
                body=str(job.response))

            self.acknowledge_message(job.basic_deliver.delivery_tag)
            log.info("Result published")
        else:
            log.info("Result not published as connection was lost")
        self._job = None

    def acknowledge_message(self, delivery_tag):
        """Acknowledge the message delivery from RabbitMQ by sending a
        Basic.Ack RPC method for the delivery tag.

        :param int delivery_tag: The delivery tag from the Basic.Deliver frame

        """
        log.info('Acknowledging message {}', delivery_tag)
        self._channel.basic_ack(delivery_tag)

    def stop_consuming(self):
        """Tell RabbitMQ that you would like to stop consuming by sending the
        Basic.Cancel RPC command.

        """
        if self._channel:
            log.info('Sending a Basic.Cancel RPC command to RabbitMQ')
            cb = functools.partial(
                self.on_cancelok, userdata=self._consumer_tag)
            self._channel.basic_cancel(self._consumer_tag, cb)

    def on_cancelok(self, _unused_frame, userdata):
        """This method is invoked by pika when RabbitMQ acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.

        :param pika.frame.Method _unused_frame: The Basic.CancelOk frame
        :param str|unicode userdata: Extra user data (consumer tag)

        """
        self._consuming = False
        log.info(
            'RabbitMQ acknowledged the cancellation of the consumer: {}',
            userdata)
        self.close_channel()

    def close_channel(self):
        """Call to close the channel with RabbitMQ cleanly by issuing the
        Channel.Close RPC command.

        """
        log.info('Closing the channel')
        self._channel.close()

    def run(self):
        """Run the task consumer by connecting to RabbitMQ and then
        starting the IOLoop to block and allow the SelectConnection to operate.

        """
        self._connection = self.connect()
        self._connection.ioloop.start()

    def stop(self):
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.

        """
        if not self._closing:
            self._closing = True
            log.info('Stopping')
            if self._consuming:
                self.stop_consuming()
                self._connection.ioloop.start()
            else:
                self._connection.ioloop.stop()
            log.info('Stopped')
            if self._job:
                self.on_job_completed(self._job)


class ReconnectingWorkerTaskConsumer(object):
    """This is a task consumer that will reconnect if the nested
    WorkerTaskConsumer indicates that a reconnect is necessary.

    """

    def __init__(self, amqp_url):
        self._reconnect_delay = 0
        self._amqp_url = amqp_url
        self._consumer = WorkerTaskConsumer(self._amqp_url)

    def run(self):
        while True:
            try:
                self._consumer.run()
            except KeyboardInterrupt:
                self._consumer.stop()
                break
            self._maybe_reconnect()

    def _maybe_reconnect(self):
        if self._consumer.should_reconnect:
            self._consumer.stop()
            reconnect_delay = self._get_reconnect_delay()
            log.info('Reconnecting after {} seconds', reconnect_delay)
            time.sleep(reconnect_delay)
            self._consumer = WorkerTaskConsumer(self._amqp_url)

    def _get_reconnect_delay(self):
        if self._consumer.was_consuming:
            self._reconnect_delay = 0
        else:
            self._reconnect_delay += 1
        if self._reconnect_delay > 30:
            self._reconnect_delay = 30
        return self._reconnect_delay


@cli.command(name="amqp-worker", hidden=True)
@click.pass_context
def amqp_worker(ctx):
    """ Run an AMQP worker """

    consumer = ReconnectingWorkerTaskConsumer(_get_url())
    log.info("Service started")
    consumer.run()


class AmqpExecutor(scheduler.NetworkExecutor):

    def __init__(self, factory, task):
        super(AmqpExecutor, self).__init__(factory)
        self.factory = factory
        self.callback_queue = None
        self.connection = None
        self.priority = config.getint("amqp", "priority", 0)
        self.task = task

    def _create_manifest(self):
        manifest = JoltManifest.export(self.task)
        build = manifest.create_build()

        tasks = [self.task.qualified_name]
        tasks += [t.qualified_name for t in self.task.extensions]

        for task in tasks:
            mt = build.create_task()
            mt.name = task

        registry = scheduler.ExecutorRegistry.get()
        for key, value in registry.get_network_parameters(self.task).items():
            param = manifest.create_parameter()
            param.key = key
            param.value = value

        routing_key = WorkerTaskConsumer.ROUTING_KEY_PREFIX
        routing_key += getattr(self.task.task, "routing_key", WorkerTaskConsumer.ROUTING_KEY_REQUEST)

        return manifest.format(), routing_key

    def _run(self, env):
        timeout = int(config.getint("amqp", "timeout", 300))
        manifest, routing_key = self._create_manifest()

        self.connect()
        self.publish_request(manifest, routing_key)

        log.debug("[AMQP] Queued {0}", self.task.short_qualified_name)

        self.task.running()
        for extension in self.task.extensions:
            extension.running()

        while self.response is None:
            try:
                self.connection.process_data_events(time_limit=timeout)
                if self.response is None:
                    self.task.info("Remote execution still in progress after {}",
                                   self.task.duration_queued)
            except (ConnectionError, AMQPConnectionError):
                log.warning("[AMQP] Lost server connection")
                self.connect()

        log.debug("[AMQP] Finished {0}", self.task.short_qualified_name)

        manifest = JoltManifest()
        with raise_task_error_on_exception(self.task, "failed to parse build result manifest"):
            manifest.parsestring(self.response)

        self.task.running(utils.duration() - float(manifest.duration))

        if manifest.result != "SUCCESS":
            output = []
            if manifest.stdout:
                output.extend(manifest.stdout.split("\n"))
            if manifest.stderr:
                output.extend(manifest.stderr.split("\n"))
            for line in output:
                log.transfer(line, self.task.identity[:8])
            for task in [self.task] + self.task.extensions:
                with task.task.report() as report:
                    remote_report = manifest.find_task(task.qualified_name)
                    if remote_report:
                        for error in remote_report.errors:
                            report.manifest.append(error)
            raise_error("[AMQP] remote build failed with status: {0}".format(manifest.result))

        raise_task_error_if(
            self.task.has_artifact() and not env.cache.is_available_remotely(self.task), self.task,
            "no task artifact available in any cache, check configuration")

        raise_task_error_if(
            self.task.has_artifact() and not env.cache.download(self.task) and env.cache.download_enabled(),
            self.task, "failed to download task artifact")

        for extension in self.task.extensions:
            raise_task_error_if(
                self.task.has_artifact() and not env.cache.download(extension) and env.cache.download_enabled(),
                self.task, "failed to download task artifact")

        return self.task

    @utils.retried.on_exception((ConnectionError, AMQPConnectionError, AMQPConnectorStackTimeout))
    def connect(self):
        self.connection = pika.BlockingConnection(
            parameters=pika.URLParameters(_get_url()))
        self.channel = self.connection.channel()
        if not self.callback_queue:
            self.corr_id = self.task.identity
            self.channel.exchange_declare(
                exchange=WorkerTaskConsumer.RESULT_EXCHANGE,
                exchange_type=WorkerTaskConsumer.EXCHANGE_TYPE)
            result = self.channel.queue_declare(
                '', arguments={"x-expires": 7200000})
            self.callback_queue = result.method.queue
            self.channel.queue_bind(
                self.callback_queue,
                WorkerTaskConsumer.RESULT_EXCHANGE,
                routing_key=self.corr_id)
        self.channel.basic_consume(
            queue=self.callback_queue,
            on_message_callback=self.on_response,
            auto_ack=False)
        log.debug("[AMQP] Established connection to server")

    def on_response(self, channel, basic_deliver, properties, body):
        # log.debug("[AMQP] Completion of {}, expecting {}", properties.correlation_id, self.corr_id)
        if self.corr_id == properties.correlation_id:
            self.response = body.decode()
            channel.basic_ack(basic_deliver.delivery_tag)
        else:
            channel.basic_ack(basic_deliver.delivery_tag)

    def publish_request(self, manifest, routing_key):
        props = pika.BasicProperties(
            correlation_id=self.corr_id,
            priority=self.priority,
            headers={"x-deduplication-header": self.task.identity})
        self.response = None
        self.channel.basic_publish(
            exchange=WorkerTaskConsumer.EXCHANGE,
            routing_key=routing_key,
            properties=props,
            body=manifest)

    def run(self, env):
        try:
            self.task.started(TYPE)
            hooks.task_started_execution(self.task)
            for extension in self.task.extensions:
                extension.started(TYPE)
                hooks.task_started_execution(extension)
            with hooks.task_run([self.task] + self.task.extensions):
                self._run(env)
            for extension in self.task.extensions:
                hooks.task_finished_execution(extension)
                extension.finished(TYPE)
            hooks.task_finished_execution(self.task)
            self.task.finished(TYPE)
        except (ConnectionError, AMQPConnectionError):
            log.exception()
            for extension in self.task.extensions:
                extension.failed(TYPE)
            self.task.failed(TYPE)
            raise_error("Lost connection to AMQP server")
        except Exception as e:
            log.exception()
            for extension in self.task.extensions:
                extension.failed(TYPE)
            self.task.failed(TYPE)
            raise e
        finally:
            if self.connection is not None:
                utils.call_and_catch(self.connection.close)
        return self.task


@scheduler.ExecutorFactory.Register
class AmqpExecutorFactory(scheduler.NetworkExecutorFactory):
    def __init__(self, options):
        workers = config.getint(NAME, "workers", 16)
        super(AmqpExecutorFactory, self).__init__(max_workers=workers)
        self._options = options

    @property
    def options(self):
        return self._options

    def create(self, task):
        return AmqpExecutor(self, task)


log.verbose("[AMQP] Loaded")



import sys

from jolt import filesystem as fs
from jolt.cache import ArtifactStringAttribute
from jolt.cache import ArtifactAttributeSet
from jolt.cache import ArtifactAttributeSetProvider


class PythonVariable(ArtifactStringAttribute):
    def __init__(self, artifact, name):
        super(PythonVariable, self).__init__(artifact, name)

    def apply(self, task, artifact):
        pass

    def unapply(self, task, artifact):
        pass


class PythonListVariable(PythonVariable):
    def __init__(self, artifact, name):
        super(PythonListVariable, self).__init__(artifact, name)

    def append(self, value):
        if self.get_value():
            self.set_value(self.get_value() + ":" + value)
        else:
            self.set_value(value)

    def items(self):
        value = self.get_value()
        return value.split(":") if value is not None else []


class PythonPathVariable(PythonListVariable):
    def __init__(self, artifact, name):
        super(PythonPathVariable, self).__init__(artifact, name)

    def append(self, value):
        if self.get_value():
            self.set_value(self.get_value() + ":" + value)
        else:
            self.set_value(value)

    def items(self):
        value = self.get_value()
        return value.split(":") if value is not None else []

    def apply(self, task, artifact):
        paths = [fs.path.join(artifact.path, path) for path in self.items()]
        sys.path.extend(paths)

    def unapply(self, task, artifact):
        paths = [fs.path.join(artifact.path, path) for path in self.items()]
        for path in paths:
            sys.path.remove(path)


class PythonDictVariable(PythonListVariable):
    def __init__(self, artifact, name):
        super(PythonDictVariable, self).__init__(artifact, name)

    def __setitem__(self, key, value=None):
        item = "{0}={1}".format(key, value) if value is not None else key
        self.append(item)


class Python(ArtifactAttributeSet):
    def __init__(self, artifact):
        super(Python, self).__init__()
        super(ArtifactAttributeSet, self).__setattr__("_artifact", artifact)

    def create(self, name):
        if name.lower() == "path":
            return PythonPathVariable(self._artifact, "PATH")
        assert False, "no such python attribute: {0}".format(name)


@ArtifactAttributeSetProvider.Register
class PythonProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "python", Python(artifact))

    def parse(self, artifact, content):
        if "python" not in content:
            return
        for key, value in content["python"].items():
            setattr(artifact.python, key, value)

    def format(self, artifact, content):
        if "python" not in content:
            content["python"] = {}
        for key, value in artifact.python.items():
            content["python"][key] = str(value)

    def apply(self, task, artifact):
        artifact.python.apply(task, artifact)

    def unapply(self, task, artifact):
        artifact.python.unapply(task, artifact)



"""
Reimplementation of Python's inspect module.

A bugfix was made in Python 3.9 which significantly slowed
down the builtin inspect module's source lookup functions.

Unlike the builtin module, this implementation caches the AST
of already parsed modules in order to speed up those functions.

"""

import ast
import sys


# Cache of parsed modules (ClassFinder objects), indexed by module.
_modules = {}


def _populate_cache(module):
    global _modules

    class _ClassFinder(ast.NodeVisitor):
        """ Collects all classes and functions in an Abstract Syntax Tree. """

        def __init__(self):
            self._stack = []
            self.classes = {}
            self.functions = {}

        def visit_FunctionDef(self, node):
            self._stack.append(node.name)
            self.functions['.'.join(self._stack)] = node
            self._stack.append('<locals>')
            self.generic_visit(node)
            self._stack.pop()
            self._stack.pop()

        visit_AsyncFunctionDef = visit_FunctionDef

        def visit_ClassDef(self, node):
            self._stack.append(node.name)
            self.classes['.'.join(self._stack)] = node
            self.generic_visit(node)
            self._stack.pop()

    with open(module.__file__) as modfp:
        tree = ast.parse(modfp.read())
    class_finder = _ClassFinder()
    class_finder.visit(tree)
    _modules[module] = class_finder
    return class_finder


def getmodule(cls_or_func):
    """ Returns module in which a class or function is defined """
    if hasattr(cls_or_func, '__module__'):
        return sys.modules.get(cls_or_func.__module__)
    raise TypeError('{!r} is a built-in object'.format(cls_or_func))


def getfile(cls_or_func):
    """ Returns the name of the file in which a class or function is defined """
    module = getmodule(cls_or_func)
    if getattr(module, '__file__', None):
        return module.__file__
    raise TypeError('{!r} is a built-in object'.format(cls_or_func))


def getlineno(cls_or_func):
    """ Returns the name of the file in which a class or function is defined """
    if isinstance(cls_or_func, type):
        ast = getclassast(cls_or_func)
    else:
        ast = getfuncast(cls_or_func)
    return ast.lineno


def getclassast(cls):
    """ Returns the Abstract Syntax Tree of a class """
    global _modules
    module = getmodule(cls)
    module_ast = _modules.get(module)
    if not module_ast:
        module_ast = _populate_cache(module)
    return module_ast.classes[cls.__qualname__]


def getfuncast(func):
    """ Returns the Abstract Syntax Tree of a function """
    global _modules
    module = getmodule(func)
    module_ast = _modules.get(module)
    if not module_ast:
        module_ast = _populate_cache(module)
    return module_ast.functions[func.__qualname__]


def getclasssource(cls):
    """ Returns the source code of a class """
    tree = getclassast(cls)
    try:
        return ast.unparse(tree)
    except AttributeError:
        # For Python <3.9
        return ast.dump(tree, annotate_fields=False)


def getfuncsource(func):
    """ Returns the source code of a function """
    tree = getfuncast(func)
    return ast.dump(tree, annotate_fields=False)


def getmoduleclasses(module, searchtypes, predicate=None):
    """
    Finds all classes in a module that are subclasses of one of
    the search types.

    Returns a dictionary of class lists, indexed by the search types.
    """
    classes = {searchtype: [] for searchtype in searchtypes}

    for key in dir(module):
        obj = getattr(module, key)

        # Skip non-classes
        if not isinstance(obj, type):
            continue

        # Check if subclass of searched types
        for searchtype in searchtypes:
            if issubclass(obj, searchtype):
                if not predicate or not predicate(obj):
                    classes[searchtype].append(obj)

    return classes



import datetime
import hashlib
import os
from pathlib import Path, PurePath

from jolt import inspection
from jolt import utils
from jolt import filesystem as fs
from jolt import tools


_providers = []


@utils.Singleton
class HashInfluenceRegistry(object):
    @staticmethod
    def Register(cls):
        _providers.append(cls)

    def __init__(self):
        self._providers = [provider() for provider in _providers]

    def register(self, provider):
        self._providers.append(provider)

    def apply_all(self, task, sha):
        for influence in self.get_strings(task):
            sha.update(influence.encode())

    def get_strings(self, task):
        content = []
        for provider in self._providers + task.influence:
            for line in str(provider.get_influence(task)).splitlines():
                content.append("Influence-{0}: {1}".format(provider.name, line))
        return content


class HashInfluenceProvider(object):
    name = "X"

    def get_influence(self, task):
        raise NotImplementedError()


class TaintInfluenceProvider(object):
    name = "Taint"

    def get_influence(self, task):
        return str(task.taint)


class TaskAttributeInfluence(HashInfluenceProvider):
    def __init__(self, attrib, sort=False):
        self._attrib = attrib
        self._sort = sort
        self.name = attrib.title()

    def get_influence(self, task):
        value = getattr(task, tools.Tools(task).expand(self._attrib), "N/A")
        try:
            value = value.__get__(task)
            if type(value) == list and self._sort:
                value.sort()
        except AttributeError:
            pass
        return value


def attribute(name, type=None, sort=False):
    """ Add task attribute value as hash influence.

    Args:
        name (str): Name of task class attribute/property.
        type (InfluenceProvider, optional): Alternative
            HashInfluenceProvider implementation used to interpret
            the value of the attribute. If the attribute is a list,
            the provider will be instantiated once for each item in
            the list. For example, if the attribute is a list of
            filesystem paths, the FileInfluence class could be passed
            as argument to automatically monitor files at those paths.
        sort (boolean, optional): Optionally sort the value. Always sort
            values that are unstable, such as dictionaries.

    Example:

    .. code-block:: python

        from jolt import influence

        @influence.attribute("attribute")
        class Example(Task):
            attribute = False

    .. code-block:: python

        from jolt import influence

        @influence.attribute("sources", type=influence.FileInfluence)
        class Example(Task):
            sources = [
                "main.cpp",
                "utils.cpp",
            ]
    """

    def _decorate(cls):
        _old_influence = cls._influence

        def _influence(self, *args, **kwargs):
            influence = _old_influence(self, *args, **kwargs)
            if type:
                items = getattr(self, name, [])
                if callable(items):
                    items = items()
                if sort:
                    items = sorted(items)
                for item in items:
                    influence.append(type(item))
            else:
                influence.append(TaskAttributeInfluence(name, sort))
            return influence

        cls._influence = _influence
        return cls
    return _decorate


class TaskSourceInfluence(HashInfluenceProvider):
    def __init__(self, funcname, obj=None):
        self.name = "Source"
        self.funcname = funcname
        self.obj = obj

    @staticmethod
    def _default_func():
        pass

    def get_influence(self, task):
        obj = self.obj or task
        try:
            funcname = self.funcname
            funcname = obj.expand(funcname)
        except Exception:
            pass

        # Collect all functions from the class hierarchy
        if type(obj) == type:
            funcs = [utils.getattr_safe(mro, funcname, TaskSourceInfluence._default_func)
                     for mro in obj.mro()]
        else:
            funcs = [utils.getattr_safe(mro, funcname, TaskSourceInfluence._default_func)
                     for mro in obj.__class__.mro()]

        # Calculate sha1 sum for all functions
        shasum = hashlib.sha1()
        for func in funcs:
            try:
                func.__influence
            except AttributeError:
                func.__influence = utils.sha1(inspection.getfuncsource(func))
            finally:
                shasum.update(func.__influence.encode())

        return shasum.hexdigest() + ": " + funcname


def source(name, obj=None):
    """ Add function source code as hash influence.

    Args:
        name (str): Name of function in task class.

    Example:

    .. code-block:: python

        from jolt import influence

        @influence.source("method")
        class Example(Task):
            def method(self):
                return False

    """
    def _decorate(cls):
        _old_influence = cls._influence

        def _influence(self, *args, **kwargs):
            influence = _old_influence(self, *args, **kwargs)
            influence.append(TaskSourceInfluence(name, obj))
            return influence

        cls._influence = _influence
        return cls
    return _decorate


class TaskClassSourceInfluence(HashInfluenceProvider):
    name = "Source"

    def get_influence(self, task):
        # Calculate sha1 sum for classes in hierarchy
        result = ""
        for cls in reversed(task.__class__.mro()):
            if cls is object:
                continue
            try:
                cls.__dict__["_TaskClassSourceInfluence__influence"]
            except KeyError:
                try:
                    cls.__influence = utils.sha1(inspection.getclasssource(cls))
                except TypeError:
                    continue
            result += cls.__dict__["_TaskClassSourceInfluence__influence"] + \
                ": " + cls.__name__ + "\n"
        return result


class TaskRequirementInfluence(HashInfluenceProvider):
    name = "Requirement"

    def __init__(self, proxy):
        self._identity = proxy.identity
        self._name = proxy.short_qualified_name

    def get_influence(self, task):
        return "{}: {}".format(self._identity, self._name)


@HashInfluenceRegistry.Register
class TaskNameInfluence(HashInfluenceProvider):
    name = "Name"

    def get_influence(self, task):
        return task.name


@HashInfluenceRegistry.Register
class TaskParameterInfluence(HashInfluenceProvider):
    name = "Parameters"

    def get_influence(self, task):
        return ",".join(
            sorted(["{0}={1}".format(key, value)
                    for key, value in task._get_parameter_objects().items()
                    if value.is_influencer()]))


class InstanceInfluence(HashInfluenceProvider):
    name = "Instance"

    def get_influence(self, task):
        return task._instance.value


def always(cls):
    """ Always execute the task.

    Example:

      .. code-block:: python

        from jolt import influence

        @influence.always
        class Example(Task):

    """
    _old_influence = cls._influence

    def _influence(self, *args, **kwargs):
        influence = _old_influence(self, *args, **kwargs)
        influence.append(InstanceInfluence())
        return influence

    cls._influence = _influence
    return cls


class TaskDateInfluence(HashInfluenceProvider):
    name = "Date"

    def __init__(self, fmt):
        self.fmt = fmt

    def get_influence(self, task):
        now = datetime.datetime.now()
        return now.strftime(self.fmt)


def _date_influence(fmt):
    def _decorate(cls):
        _old_influence = cls._influence

        def _influence(self, *args, **kwargs):
            influence = _old_influence(self, *args, **kwargs)
            influence.append(TaskDateInfluence(fmt))
            return influence

        cls._influence = _influence
        return cls
    return _decorate


yearly = _date_influence("%Y")
""" Add yearly hash influence.

If nothing else changes, the task is re-executed once every year.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.yearly
        class Example(Task):

"""


weekly = _date_influence("%Y-%w")
""" Add weekly hash influence.

If nothing else changes, the task is re-executed once every week.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.weekly
        class Example(Task):

"""


monthly = _date_influence("%Y-%m")
""" Add monthly hash influence.

If nothing else changes, the task is re-executed once every month.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.monthly
        class Example(Task):

"""


daily = _date_influence("%Y-%m-%d")
""" Add daily hash influence.

If nothing else changes, the task is re-executed once every day.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.daily
        class Example(Task):

"""


hourly = _date_influence("%Y-%m-%d %H")
""" Add hourly hash influence.

If nothing else changes, the task is re-executed once every hour.

Example:

    .. code-block:: python

        from jolt import influence

        @influence.hourly
        class Example(Task):

"""


class TaskEnvironmentInfluence(HashInfluenceProvider):
    name = "Environment"

    def __init__(self, variable):
        self.variable = variable

    def get_influence(self, task):
        return self.variable + "=" + os.environ.get(self.variable, "<unset>")


def environ(variable):
    """ Add environment variable hash influence.

    Args:
        variable (str): Name of an environment variable that will
            influence the hash of the task.

    Example:

    .. code-block:: python

        from jolt import influence

        @influence.environ("CFLAGS")
        class Example(Task):

    """
    def _decorate(cls):
        _old_influence = cls._influence

        def _influence(self, *args, **kwargs):
            influence = _old_influence(self, *args, **kwargs)
            influence.append(TaskEnvironmentInfluence(variable))
            return influence

        cls._influence = _influence
        return cls

    return _decorate


_fi_files = {}


class FileInfluence(HashInfluenceProvider):
    def __init__(self, path):
        self.path = path.rstrip(fs.sep)
        self.name = "File"
        self._files = {}

    def get_file_influence(self, path):
        return utils.filesha1(str(path))

    def get_filelist(self, task):
        try:
            return self._files[task]
        except KeyError:
            if fs.path.isdir(task.tools.expand_path(self.path)):
                path = self.path + fs.sep + "**"
            else:
                path = self.path

            filelist = task.tools.glob(path, expand=True)
            filelist.sort()
            filelist = [Path(fname) for fname in filelist]
            self._files[task] = filelist
            return filelist

    def get_influence(self, task):
        result = []
        for f in self.get_filelist(task):
            if f.is_dir():
                continue
            value = _fi_files.get(f)
            if value:
                result.append(value)
            elif f.exists():
                _fi_files[f] = value = self.get_file_influence(f) + ": " + f.name
                result.append(value)
            elif fs.path.lexists(str(f)):
                _fi_files[f] = value = "Symlink (broken): " + f.name
                result.append(value)
        return "\n".join(result)

    def is_influenced_by(self, task, path):
        """
        Return True if the path influences the task.
        """
        path = task.tools.expand_path(path)
        return PurePath(path) in self.get_filelist(task)


class DirectoryInfluence(FileInfluence):
    def __init__(self, path):
        super().__init__(path.rstrip(os.sep) + "/**")


def files(pathname):
    """ Add file content hash influence.

    Args:
        pathname (str): A pathname pattern used to find files that will
                influence the hash of the task
                The pattern may contain simple shell-style
                wildcards such as '*' and '?'. Note: files starting with a
                dot are not matched by these wildcards.

    Example:

    .. code-block:: python

        from jolt import influence

        @influence.files("*.cpp")
        class Example(Task):

    """
    def _decorate(cls):
        _old_influence = cls._influence

        def _influence(self, *args, **kwargs):
            influence = _old_influence(self, *args, **kwargs)
            influence.append(FileInfluence(pathname))
            return influence

        cls._influence = _influence
        return cls

    return _decorate


def global_files(pathname, cls=FileInfluence):
    HashInfluenceRegistry.get().register(cls(pathname))


class WhitelistInfluence(FileInfluence):
    def __init__(self, path):
        self.path = path.rstrip(fs.sep)
        self.name = "Whitelist"

    def get_influence(self, task):
        return self.path

    def is_influenced_by(self, task, path):
        path = task.tools.expand_path(path)
        pattern = task.tools.expand_path(self.path)
        return utils.pathmatch(path, pattern)


def whitelist(pathname):
    """
    Whitelist files published by a task, but don't let them influence
    the hash.

    This is typically used to whitelist files in the workspace. It
    should only be used for files produced by the task, not for files
    used as input. If possible, consider writing files to build
    directories instead.

    This decorator cannot be used to exclude influencing files matched
    by another influence decorator. If there is an overlap of files
    between decorators, some or all of the overlapping files files may
    influence the hash.

    Args:
        pathname (str): A pathname pattern used to match whitelisted files.
                The pattern may contain simple shell-style
                wildcards such as '*' and '?'. Note: files starting with a
                dot are not matched by these wildcards.

    Example:

    .. code-block:: python

        from jolt import influence

        @influence.whitelist("build/")
        class Example(Task):

    """
    def _decorate(cls):
        _old_influence = cls._influence

        def _influence(self, *args, **kwargs):
            influence = _old_influence(self, *args, **kwargs)
            influence.append(WhitelistInfluence(pathname))
            return influence

        cls._influence = _influence
        return cls

    return _decorate


class StringInfluence(HashInfluenceProvider):
    name = "String"

    def __init__(self, value):
        self.value = value

    def get_influence(self, task):
        return self.value


def global_string(string):
    HashInfluenceRegistry.get().register(StringInfluence(string))



from jolt import influence
from jolt.plugins import ninja, yamltask


@influence.attribute("yaml")
class LibraryYamlTask(ninja.CXXLibrary):
    def run(self, d, t):
        super(ninja.CXXLibrary, self).run(d, t)
        yamltask.YamlTask.run(self, d, t)

    def publish(self, a, t):
        super(ninja.CXXLibrary, self).publish(a, t)
        yamltask.YamlTask.publish(self, a, t)


@yamltask.register
class LibraryYamlTaskBuilder(yamltask.YamlTaskBuilder):
    name = "library"

    def build(self, ytask):
        class LoadedYamlTask(LibraryYamlTask):
            # From Task
            name = None
            requires = []
            cacheable = True
            commands = []
            collects = []
            extends = ""
            fast = False
            influence = []
            joltdir = "."
            joltproject = None
            metadata = {}
            selfsustained = False
            taint = False
            weight = False

            # From CXXLibrary
            asflags = []
            cflags = []
            cxxflags = []
            depimports = []
            incpaths = []
            libpaths = []
            libraries = []
            ldflags = []
            macros = []
            sources = []
            publishdir = "lib/"
            shared = False
            source_influence = True
            binary = None
            incremental = True

        for key, obj in ytask.items():
            setattr(LoadedYamlTask, key, obj)

        LoadedYamlTask.yaml = ytask

        return LoadedYamlTask


@influence.attribute("yaml")
class ExecutableYamlTask(ninja.CXXExecutable):
    def run(self, d, t):
        super(ninja.CXXExecutable, self).run(d, t)
        yamltask.YamlTask.run(self, d, t)

    def publish(self, a, t):
        super(ninja.CXXExecutable, self).publish(a, t)
        yamltask.YamlTask.publish(self, a, t)


@yamltask.register
class ExecutableYamlTaskBuilder(yamltask.YamlTaskBuilder):
    name = "executable"

    def build(self, ytask):
        class LoadedYamlTask(ExecutableYamlTask):
            # From Task
            name = None
            requires = []
            cacheable = True
            commands = []
            collects = []
            extends = ""
            fast = False
            influence = []
            joltdir = "."
            joltproject = None
            metadata = {}
            selfsustained = False
            taint = False
            weight = False

            # From CXXExecutable
            asflags = []
            cflags = []
            cxxflags = []
            depimports = []
            incpaths = []
            libpaths = []
            libraries = []
            ldflags = []
            macros = []
            sources = []
            publishdir = "lib/"
            shared = False
            source_influence = True
            binary = None
            incremental = True

        for key, obj in ytask.items():
            setattr(LoadedYamlTask, key, obj)

        LoadedYamlTask.yaml = ytask

        return LoadedYamlTask



import operator
import re


class version(object):
    def __init__(self, verstr):
        if type(verstr) == str:
            match = re.search(r"(?P<major>[\d]+)\.(?P<minor>[\d]+)(\.(?P<patch>[\d]+))?", verstr)
            if not match:
                raise ValueError(verstr)
            values = match.groupdict()
            self.major = int(values["major"])
            self.minor = int(values["minor"])
            self.patch = int(values["patch"]) if values["patch"] else None
        elif type(verstr) == tuple:
            if len(verstr) < 2 or len(verstr) > 3:
                raise ValueError(verstr)
            self.major = verstr[0]
            self.minor = verstr[1]
            self.patch = verstr[2] if len(verstr) == 3 else None
        else:
            raise ValueError(verstr)

    def __str__(self):
        if self.patch is None:
            return f"{self.major}.{self.minor}"
        else:
            return f"{self.major}.{self.minor}.{self.patch}"

    def __repr__(self):
        return str(self)

    def __eq__(self, version):
        if self.major != version.major:
            return False
        if self.minor != version.minor:
            return False
        if self.patch is None or version.patch is None:
            return True
        if self.patch != version.patch:
            return False
        return True

    def __lt__(self, version):
        if self.major < version.major:
            return True
        if self.major > version.major:
            return False
        if self.minor < version.minor:
            return True
        if self.minor > version.minor:
            return False
        if self.patch is None or version.patch is None:
            return False
        if self.patch < version.patch:
            return True
        return False

    def __le__(self, version):
        return self < version or self == version

    def __gt__(self, version):
        if self.major > version.major:
            return True
        if self.major < version.major:
            return False
        if self.minor > version.minor:
            return True
        if self.minor < version.minor:
            return False
        if self.patch is None or version.patch is None:
            return False
        if self.patch > version.patch:
            return True
        return False

    def __ge__(self, version):
        return self > version or self == version


class version_operator(object):
    def __init__(self, verstr):
        match = re.search(r"(?P<operator>((>|<)?=|>))?", verstr)
        if not match:
            raise ValueError(verstr)
        self.opstr = match.groupdict("=")["operator"]
        if self.opstr == "=":
            self.op = operator.eq
        elif self.opstr == ">":
            self.op = operator.gt
        elif self.opstr == ">=":
            self.op = operator.ge
        elif self.opstr == "<=":
            self.op = operator.le
        else:
            ValueError(self.opstr)

    def __call__(self, a, b):
        return self.op(a, b)

    def __repr__(self):
        return repr(self.opstr)

    def __str__(self):
        return str(self.opstr)


class requirement(object):
    def __init__(self, verstr):
        self.operator = version_operator(verstr)
        self.version = version(verstr)

    def satisfied(self, version):
        """ Returns True if the version satisfies the requirement """
        return self.operator(version, self.version)

    def required(self):
        """ Returns the least required version """
        if self.operator.opstr == ">":
            if self.version.patch is None:
                return version((self.version.major, self.version.minor + 1))
            return version((self.version.major, self.version.minor, self.version.patch + 1))
        else:
            return self.version

    def __str__(self):
        return f"{self.operator}{self.version}"



from jolt.tasks import Task, TaskRegistry
from jolt.cache import ArtifactCache
from jolt.graph import GraphBuilder
from jolt.error import raise_error_if
from jolt.manifest import JoltManifest
from jolt.scheduler import JoltEnvironment
from jolt.scheduler import LocalExecutor
from jolt.scheduler import LocalExecutorFactory
from jolt.scheduler import NetworkExecutorExtension
from jolt.scheduler import NetworkExecutorExtensionFactory
from jolt.loader import JoltLoader
from jolt import config
from jolt import filesystem as fs
from jolt import influence
from jolt import log
from jolt import utils


log.verbose("[SelfDeploy] Loaded")

_path = fs.path.dirname(__file__)
_path = fs.path.dirname(_path)
_path = fs.path.dirname(_path)


@influence.files(fs.path.join(_path, "**", "*.py"))
@influence.files(fs.path.join(_path, "**", "*.sh"))
@influence.files(fs.path.join(_path, "**", "*.xslt"))
@influence.files(fs.path.join(_path, "**", "*.template"))
class Jolt(Task):
    name = "jolt"

    def __init__(self, *args, **kwargs):
        super(Jolt, self).__init__(*args, **kwargs)
        self.influence.append(
            influence.StringInfluence(
                config.get("selfdeploy", "requires", "")))
        for e in self.extra_files:
            self.influence.append(influence.DirectoryInfluence(e))

    def _verify_influence(self, *args, **kwargs):
        pass  # FIXME: Validation doesn't work properly for directories

    @property
    def loaderdir(self):
        return JoltLoader.get().joltdir

    @property
    def extra_files(self):
        ext = config.get("selfdeploy", "extra", "")
        return [fs.path.join(self.loaderdir, e) for e in ext.split(",")] if ext else []

    @property
    def extra_dependencies(self):
        req = config.get("selfdeploy", "requires", "")
        return req.split() if req else []

    def info(self, fmt, *args, **kwargs):
        log.verbose(fmt, *args, **kwargs)

    @property
    def dependencies(self):
        """
        Queries local pip database for packages and versions used by jolt.

        The data is used to pin dependencies at specific versions so that
        the virtalenv installed on a worker matches what the client had
        installed locally.

        If pip is not available or its internal implementation has changed,
        no version pinning will be performed. Instead, workers will install
        Jolt with its default loose version requirements.
        """
        def get_installed_distributions():
            try:
                from pip._internal.metadata import get_environment
            except ImportError:
                from pip._internal.utils import misc
                return {
                    dist.project_name.lower(): dist
                    for dist in misc.get_installed_distributions()
                }
            else:
                dists = get_environment(None).iter_installed_distributions()
                return {dist._dist.project_name.lower(): dist._dist for dist in dists}

        dists = get_installed_distributions()
        reqs = ["jolt"] + [dep.lower() for dep in self.extra_dependencies]
        pkgs = {}

        while reqs:
            req = reqs.pop()

            dist = dists.get(req)
            if dist is None:
                self.info("[SelfDeploy] Dependency not found: {}", req)
                pkgs["req"] = req
                continue

            for dep in dist.requires():
                name = dep.project_name.lower()
                if name not in pkgs:
                    reqs.append(name)

            pkgs[req] = f"{dist.project_name}=={dist.version}"

        del pkgs["jolt"]
        return pkgs.values()

    def publish(self, artifact, tools):
        with tools.cwd(tools.builddir()):
            try:
                pinned_reqs = self.dependencies
                if pinned_reqs:
                    tools.write_file("requirements.txt", "\n".join(pinned_reqs))
                    artifact.collect("requirements.txt")
            except Exception:
                log.exception()
        with tools.cwd(_path):
            artifact.collect('README.rst')
            artifact.collect('setup.py')
            artifact.collect('jolt/*.py')
            artifact.collect('jolt/*.sh')
            artifact.collect('jolt/*/*.py')
            artifact.collect('jolt/*/*/*.py')
            artifact.collect('jolt/*/*.xslt')
            artifact.collect('jolt/*/*.template')
            artifact.collect('jolt/plugins/selfdeploy/README.rst', flatten=True)
            artifact.collect('jolt/plugins/selfdeploy/setup.py', flatten=True)
            for e in self.extra_files:
                with tools.cwd(fs.path.dirname(e)):
                    artifact.collect(fs.path.basename(e))


class SelfDeployExtension(NetworkExecutorExtension):
    @utils.cached.instance
    def get_parameters(self, task):
        registry = TaskRegistry()
        registry.add_task_class(Jolt)
        acache = ArtifactCache.get()
        env = JoltEnvironment(cache=acache)
        gb = GraphBuilder(registry, JoltManifest())
        dag = gb.build(["jolt"])
        task = dag.select(lambda graph, task: True)
        assert len(task) == 1, "too many selfdeploy tasks found"
        task = task[0]
        if not acache.is_available_remotely(task):
            factory = LocalExecutorFactory()
            executor = LocalExecutor(factory, task, force_upload=True)
            executor.run(env)
        jolt_url = acache.location(task)
        raise_error_if(not jolt_url, "failed to deploy jolt to a remote cache")
        return {
            "jolt_url": jolt_url,
            "jolt_identity": task.identity[:8],
            "jolt_requires": config.get("selfdeploy", "requires", "")
        }


@NetworkExecutorExtensionFactory.Register
class SelfDeployExtensionFactory(NetworkExecutorExtensionFactory):
    def create(self):
        return SelfDeployExtension()



from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.hooks import TaskHook, TaskHookFactory


log.verbose("[AutoWeight] Loaded")


class WeightHooks(TaskHook):
    def __init__(self):
        self._tasks = {}
        self._samples = config.getint("autoweight", "samples", 10)
        self.load()

    @property
    def dbpath(self):
        return fs.path.join(config.get_cachedir(), "autoweight.json")

    def load(self):
        self._tasks = utils.fromjson(self.dbpath, ignore_errors=True)

    def save(self):
        utils.tojson(self.dbpath, self._tasks, ignore_errors=True)

    def task_created(self, task):
        data = self._tasks.get(task.qualified_name)
        if data:
            task.weight = max(data)

    def task_finished(self, task):
        data = self._tasks.get(task.qualified_name)
        if data:
            data.append(task.duration_running.seconds)
            self._tasks[task.qualified_name] = data[-self._samples:]
        else:
            self._tasks[task.qualified_name] = [task.duration_running.seconds]
        self.save()


@TaskHookFactory.register
class WeightFactory(TaskHookFactory):
    def create(self, env):
        return WeightHooks()



from concurrent.futures import ThreadPoolExecutor, as_completed, Future
import os
import queue

from jolt import config
from jolt import hooks
from jolt import log
from jolt import utils
from jolt import tools
from jolt.error import raise_error
from jolt.error import raise_task_error
from jolt.error import raise_task_error_if
from jolt.graph import PruneStrategy
from jolt.manifest import ManifestExtension
from jolt.manifest import ManifestExtensionRegistry
from jolt.options import JoltOptions


class JoltEnvironment(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TaskQueue(object):
    def __init__(self, strategy):
        self.futures = {}
        self.strategy = strategy
        self.duration_acc = utils.duration_diff(0)
        self._aborted = False

    def submit(self, cache, task):
        if self._aborted:
            return None

        env = JoltEnvironment(cache=cache)
        executor = self.strategy.create_executor(task)
        raise_task_error_if(
            not executor, task,
            "no executor can execute the task; "
            "requesting a distributed network build without proper configuration?")

        task.set_in_progress()
        future = executor.submit(env)
        self.futures[future] = task
        return future

    def wait(self):
        for future in as_completed(self.futures):
            task = self.futures[future]
            try:
                future.result()
            except Exception as error:
                log.exception()
                return task, error
            finally:
                self.duration_acc += task.duration_running or 0
                del self.futures[future]
            return task, None
        return None, None

    def abort(self):
        self._aborted = True
        for future, task in self.futures.items():
            future.cancel()
        if len(self.futures):
            log.info("Waiting for tasks to finish, please be patient")
        self.strategy.executors.shutdown()

    def is_aborted(self):
        return self._aborted

    def in_progress(self, task):
        return task in self.futures.values()


class Executor(object):
    def __init__(self, factory):
        self.factory = factory

    def submit(self, env):
        return self.factory.submit(self, env)

    def is_aborted(self):
        return self.factory.is_aborted()

    def run(self, env):
        pass


class LocalExecutor(Executor):
    def __init__(self, factory, task, force_upload=False, force_build=False):
        super(LocalExecutor, self).__init__(factory)
        self.task = task
        self.force_build = force_build
        self.force_upload = force_upload

    def run(self, env):
        if self.is_aborted():
            return
        try:
            self.task.started()
            hooks.task_started_execution(self.task)
            with hooks.task_run(self.task):
                self.task.run(
                    env.cache,
                    force_build=self.force_build,
                    force_upload=self.force_upload)
        except Exception as e:
            log.exception()
            self.task.failed()
            raise e
        else:
            hooks.task_finished_execution(self.task)
            self.task.finished()
        return self.task


class NetworkExecutor(Executor):
    pass


class SkipTask(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super(SkipTask, self).__init__(factory, *args, **kwargs)
        self.task = task

    def run(self, env):
        self.task.skipped()
        for ext in self.task.extensions:
            ext.skipped()
        return self.task


class Downloader(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super(Downloader, self).__init__(factory, *args, **kwargs)
        self.task = task

    def _download(self, env, task):
        if self.is_aborted():
            return
        if not task.is_downloadable():
            return
        try:
            task.started("Download")
            hooks.task_started_download(task)
            raise_task_error_if(
                not env.cache.download(task),
                task, "Failed to download task artifact")
        except Exception as e:
            with task.task.report() as report:
                report.add_exception(e)
            task.failed("Download")
            raise e
        else:
            hooks.task_finished_download(task)
            task.finished("Download")

    def run(self, env):
        self._download(env, self.task)
        for ext in self.task.extensions:
            self._download(env, ext)
        return self.task


class Uploader(Executor):
    def __init__(self, factory, task, *args, **kwargs):
        super(Uploader, self).__init__(factory, *args, **kwargs)
        self.task = task

    def _upload(self, env, task):
        if self.is_aborted():
            return
        try:
            task.started("Upload")
            hooks.task_started_upload(task)
            raise_task_error_if(
                not env.cache.upload(task),
                task, "Failed to upload task artifact")
        except Exception as e:
            with task.task.report() as report:
                report.add_exception(e)
            task.failed("Upload")
            raise e
        else:
            hooks.task_finished_upload(task)
            task.finished("Upload")

    def run(self, env):
        self._upload(env, self.task)
        for ext in self.task.extensions:
            self._upload(env, ext)

        return self.task


@utils.Singleton
class ExecutorRegistry(object):
    executor_factories = []
    extension_factories = []

    def __init__(self, options=None):
        self._options = options or JoltOptions()
        self._factories = [factory(self._options) for factory in self.__class__.executor_factories]
        self._local_factory = LocalExecutorFactory(self._options)
        self._concurrent_factory = ConcurrentLocalExecutorFactory(self._options)
        self._extensions = [factory().create() for factory in self.__class__.extension_factories]

    def shutdown(self):
        for factory in self._factories:
            factory.shutdown()
        self._local_factory.shutdown()
        self._concurrent_factory.shutdown()

    def create_skipper(self, task):
        return SkipTask(self._concurrent_factory, task)

    def create_downloader(self, task):
        # TODO: Switch to concurrent factory once the progress bar can handle it
        return Downloader(self._concurrent_factory, task)

    def create_uploader(self, task):
        # TODO: Switch to concurrent factory once the progress bar can handle it
        return Uploader(self._concurrent_factory, task)

    def create_local(self, task, force=False):
        task.set_locally_executed()
        return self._local_factory.create(task, force=force)

    def create_network(self, task):
        for factory in self._factories:
            executor = factory.create(task)
            if executor is not None:
                task.set_remotely_executed()
                return executor
        return self.create_local(task)

    def get_network_parameters(self, task):
        parameters = {}
        for extension in self._extensions:
            parameters.update(extension.get_parameters(task))
        return parameters


class NetworkExecutorExtensionFactory(object):
    @staticmethod
    def Register(cls):
        ExecutorRegistry.extension_factories.insert(0, cls)
        return cls

    def create(self):
        raise NotImplementedError()


class NetworkExecutorExtension(object):
    def get_parameters(self, task):
        return {}


class Job(object):
    def __init__(self, priority, future, executor, env):
        self.priority = priority
        self.future = future
        self.executor = executor
        self.env = env

    def __le__(self, o):
        return self.priority <= o.priority

    def __ge__(self, o):
        return self.priority >= o.priority

    def __lt__(self, o):
        return self.priority < o.priority

    def __gt__(self, o):
        return self.priority > o.priority

    def __eq__(self, o):
        return self.priority == o.priority


class ExecutorFactory(object):
    @staticmethod
    def Register(cls):
        ExecutorRegistry.executor_factories.insert(0, cls)
        return cls

    def __init__(self, options=None, max_workers=None):
        self.pool = ThreadPoolExecutor(max_workers=max_workers)
        self._aborted = False
        self._queue = queue.PriorityQueue()
        self._options = options or JoltOptions()

    def is_aborted(self):
        return self._aborted

    def is_keep_going(self):
        return self._options.keep_going

    def shutdown(self):
        self._aborted = True
        self.pool.shutdown()

    def create(self, task):
        raise NotImplementedError()

    def _run(self):
        job = self._queue.get(False)
        self._queue.task_done()
        try:
            if not self.is_aborted():
                job.executor.run(job.env)
        except KeyboardInterrupt as e:
            raise_error("Interrupted by user")
            self._aborted = True
            job.future.set_exception(e)
        except Exception as e:
            if not self.is_keep_going():
                self._aborted = True
            job.future.set_exception(e)
        else:
            job.future.set_result(job.executor)

    def submit(self, executor, env):
        future = Future()
        self._queue.put(Job(-executor.task.weight, future, executor, env))
        self.pool.submit(self._run)
        return future


class LocalExecutorFactory(ExecutorFactory):
    def __init__(self, options=None):
        max_workers = config.getint(
            "jolt", "parallel_tasks",
            os.getenv("JOLT_PARALLEL_TASKS", 1 if options is None else options.jobs))
        super(LocalExecutorFactory, self).__init__(
            options=options,
            max_workers=max_workers)

    def create(self, task, force=False):
        return LocalExecutor(self, task, force_build=force)


class ConcurrentLocalExecutorFactory(ExecutorFactory):
    def __init__(self, options=None):
        max_workers = tools.Tools().thread_count()
        super(ConcurrentLocalExecutorFactory, self).__init__(
            options=options,
            max_workers=max_workers)

    def create(self, task):
        raise NotImplementedError()


class NetworkExecutorFactory(ExecutorFactory):
    def __init__(self, *args, **kwargs):
        super(NetworkExecutorFactory, self).__init__(*args, **kwargs)


class ExecutionStrategy(object):
    def create_executor(self, task):
        raise NotImplementedError()


class LocalStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_alias():
            return self.executors.create_skipper(task)
        if not task.is_cacheable():
            return self.executors.create_local(task)
        if task.is_available_locally(self.cache):
            return self.executors.create_skipper(task)
        if self.cache.download_enabled() and task.is_available_remotely(self.cache):
            return self.executors.create_downloader(task)
        return self.executors.create_local(task)

    def should_prune_requirements(self, task):
        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_locally(self.cache):
            return True
        if self.cache.download_enabled() and task.is_available_remotely(self.cache):
            return True
        return False


class DownloadStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_alias():
            return self.executors.create_skipper(task)
        if task.is_resource():
            return self.executors.create_local(task)
        if not task.is_cacheable():
            return self.executors.create_skipper(task)
        if task.is_available_locally(self.cache):
            return self.executors.create_skipper(task)
        if self.cache.download_enabled() and task.is_available_remotely(self.cache):
            return self.executors.create_downloader(task)
        raise_task_error(task, "task must be built first")

    def should_prune_requirements(self, task):
        return False


class DistributedStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_alias():
            return self.executors.create_skipper(task)

        if task.is_resource():
            if task.deps_available_locally(self.cache):
                return self.executors.create_local(task)
            else:
                return self.executors.create_skipper(task)

        if not task.is_cacheable():
            return self.executors.create_network(task)

        if not self.cache.upload_enabled():
            return self.executors.create_network(task)

        if not task.is_goal(with_extensions=False):
            task.disable_download()
        for extension in task.extensions:
            if not extension.is_goal(with_extensions=False):
                extension.disable_download()

        remote = task.is_available_remotely(self.cache)
        if remote:
            if task.is_goal() and self.cache.download_enabled() and \
               not task.is_available_locally(self.cache):
                return self.executors.create_downloader(task)
            return self.executors.create_skipper(task)
        else:
            if task.is_available_locally(self.cache) and task.is_uploadable(self.cache):
                return self.executors.create_uploader(task)
            if task.is_fast() and task.deps_available_locally(self.cache):
                return self.executors.create_local(task)

        return self.executors.create_network(task)

    def should_prune_requirements(self, task):
        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_remotely(self.cache):
            return True
        return False


class WorkerStrategy(ExecutionStrategy, PruneStrategy):
    def __init__(self, executors, cache):
        self.executors = executors
        self.cache = cache

    def create_executor(self, task):
        if task.is_resource():
            return self.executors.create_local(task)

        if task.is_alias():
            return self.executors.create_skipper(task)

        raise_task_error_if(
            not self.cache.upload_enabled(), task,
            "artifact upload must be enabled for workers, fix configuration")

        if not task.is_cacheable():
            return self.executors.create_local(task)

        if task.is_available_locally(self.cache):
            if task.is_goal() and not task.is_available_remotely(self.cache):
                # Unpacked artifacts may become unpacked before we manage to upload.
                # To keep the implementation simple we take the easy road and rebuild
                # all artifacts that have not been unpacked, even if they are uploadable.
                if task.is_unpacked(self.cache) and task.is_uploadable(self.cache):
                    return self.executors.create_uploader(task)
                else:
                    return self.executors.create_local(task, force=True)
            return self.executors.create_skipper(task)

        if not self.cache.download_enabled():
            return self.executors.create_local(task)

        if task.is_available_remotely(self.cache):
            return self.executors.create_downloader(task)

        return self.executors.create_local(task)

    def should_prune_requirements(self, task):
        if task.is_alias() or not task.is_cacheable():
            return False
        if task.is_available_locally(self.cache):
            # Unpacked artifacts may become unpacked before we manage to upload.
            # To keep the implementation simple we take the easy road and rebuild
            # all artifacts that have not been unpacked, even if they are uploadable.
            if task.is_unpacked(self.cache) and task.is_uploadable(self.cache):
                return True
        if not task.is_goal() and task.task.selfsustained:
            return True
        return False


class TaskIdentityExtension(ManifestExtension):
    def export_manifest(self, manifest, task):
        for child in [task] + task.extensions + task.descendants:
            manifest_task = manifest.find_task(child.qualified_name)
            if manifest_task is None:
                manifest_task = manifest.create_task()
                manifest_task.name = child.qualified_name
            manifest_task.identity = child.identity


ManifestExtensionRegistry.add(TaskIdentityExtension())


class TaskExportExtension(ManifestExtension):
    def export_manifest(self, manifest, task):
        short_task_names = set()
        for child in [task] + task.extensions + task.descendants:
            manifest_task = manifest.find_task(child.qualified_name)
            if manifest_task is None:
                manifest_task = manifest.create_task()
                manifest_task.name = child.qualified_name
            for key, export in child.task._get_export_objects().items():
                attrib = manifest_task.create_attribute()
                attrib.name = key
                attrib.value = export.export(child.task)
            short_task_names.add(child.name)

        # Figure out if any task with an overridden default parameter
        # value was included in the manifest. If so, add info about it.
        default_task_names = set()
        for task in task.options.default:
            short_name, _ = utils.parse_task_name(task)
            if short_name in short_task_names:
                default_task_names.add(task)
        if default_task_names:
            build = manifest.create_build()
            for task in default_task_names:
                default = build.create_default()
                default.name = task


ManifestExtensionRegistry.add(TaskExportExtension())



import os
import errno
import ntpath
import pathlib
import posixpath
import shutil
import tempfile
import sys

from jolt.error import raise_error_if


path = os.path
sep = os.sep
anysep = [posixpath.sep, ntpath.sep]
pathsep = os.pathsep


def as_posix(path):
    return pathlib.Path(path).as_posix()


def as_dirpath(path):
    return path if path[-1] == sep else path + sep


def is_relative_to(pathname, rootdir):
    try:
        pathlib.Path(pathname).relative_to(pathlib.Path(rootdir))
        return True
    except ValueError:
        return False


def userhome():
    return os.path.expanduser("~")


def makedirs(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def mkdir(path):
    try:
        os.mkdir(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


mkdtemp = tempfile.mkdtemp


def exists(path):
    return os.path.exists(path)


def identical_files(path1, path2):
    stat1 = os.stat(path1)
    stat2 = os.stat(path2)
    if stat1.st_size != stat2.st_size:
        return False
    with open(path1, "rb") as f1, open(path2, "rb") as f2:
        for data1, data2 in iter(lambda: (f1.read(0x10000), f2.read(0x10000)), (b'', b'')):
            if data1 != data2:
                return False
    return True


def rename(old, new):
    return os.rename(old, new)


def move(src, dst):
    return shutil.move(src, dst)


def rmtree(path, ignore_errors=False):
    def onerror(func, path, exc_info):
        if os.path.isdir(path):
            try:
                os.rmdir(path)
            except Exception:
                pass
            else:
                return
        if not ignore_errors:
            _, exc, _ = exc_info
            raise exc

    shutil.rmtree(path, onerror=onerror)


def unlink(path, ignore_errors=False, tree=False):
    try:
        if os.path.isdir(path) and not os.path.islink(path):
            if tree:
                rmtree(path, ignore_errors=ignore_errors)
            else:
                os.rmdir(path)
        else:
            os.unlink(path)
    except Exception as e:
        if not ignore_errors:
            raise e


_symlinks = None


def has_symlinks():
    global _symlinks
    if _symlinks is None:
        if os.name != "nt" or (sys.getwindowsversion().major >= 10 and sys.version_info.major >= 3 and sys.version_info.minor >= 8):
            _symlinks = True
        else:
            _symlinks = False
    return _symlinks


def symlink(src, dest, *args, **kwargs):
    if os.name == "nt":
        # Try to use junctions first.
        try:
            import _winapi
            _winapi.CreateJunction(src, dest)
            return
        except KeyboardInterrupt as e:
            raise e
        except Exception:
            # Ok, probably linking a file and not a directory
            # trying a regular symlink.
            try:
                os.symlink(src, dest, *args, **kwargs)
            except OSError as e:
                raise_error_if(
                    "symbolic link privilege not held" in str(e),
                    "Permission denied while attempting to create a symlink: {}\n"
                    "Please ensure the 'Create symbolic links' right is granted "
                    "to your user in the 'Local Security Policy'.", dest)
                raise e
    else:
        os.symlink(src, dest, *args, **kwargs)


def _copy_symlink(src, dst):
    if os.path.lexists(dst):
        unlink(dst, ignore_errors=True)
    if os.path.islink(src):
        return symlink(os.readlink(src), dst)
    return shutil.copy(src, dst)


def _copy2_symlink(src, dst):
    if os.path.lexists(dst):
        unlink(dst, ignore_errors=True)
    if os.path.islink(src):
        symlink(os.readlink(src), dst)
        shutil.copystat(src, dst, follow_symlinks=False)
        return
    return shutil.copy2(src, dst)


def copy(src, dst, symlinks=False, ignore=None, metadata=True):
    dstdir = os.path.dirname(dst)
    if not os.path.isdir(dstdir):
        unlink(dstdir, ignore_errors=True)
        makedirs(dstdir)

    if symlinks:
        copyfn = _copy2_symlink if metadata else _copy_symlink
    else:
        copyfn = shutil.copy2 if metadata else shutil.copy

    if symlinks and os.path.islink(src):
        return copyfn(src, dst)
    elif not os.path.isdir(src):
        return copyfn(src, dst)

    try:
        return shutil.copytree(
            src, dst,
            symlinks,
            dirs_exist_ok=True,
            copy_function=copyfn,
        )
    except TypeError:
        return shutil.copytree(
            src, dst,
            symlinks,
            copy_function=copyfn,
        )


def scandir(scanpath, filterfn=lambda path: path[0] != ".", relative=False):
    def relresult(path, fp):
        return os.path.relpath(os.path.join(path, fp), scanpath)

    def absresult(path, fp):
        return os.path.join(path, fp)

    resfn = relresult if relative else absresult
    return [resfn(path, f)
            for path, dirs, files in os.walk(scanpath)
            for f in files
            if filterfn(f)]


def get_archive(path):
    return path + ".tar.gz"
